/* UpTools v8.6
 *
 * Copyright (c) 2005-2011 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// UpSelect<ChannelPtrClass> template checks events in any of its configured
// channels. A channel is a structure or class associated with a socket, a file,// fifo, etc. If no ChannelPtrClass is specified it defaults to (void*)

// Multithreading: UpSelect should not be accessed by several threads. If used
// in a multithreading application, it should be used by only one of the threads

//////////////////////////////////////////////////////////////////////////
// An example usage could be to handle files and  individual buffers
// and state associated with each file:
//
// class myChanelClass {
//		private:
//			int fileDescriptor, byteCount;
//			char buf[1024];
//		public:
//			...  // various class methods
// };
// ...
// UpSelect<myChannelClass*> mySelect;
// myChannelClass channel1, channel2, *cptr;
// ...
// mySelect.set(channel1.fileDescriptor, &channel1, true, false, false);
// mySelect.set(channel2.fileDescriptor, &channel2, true, false, false);
// mySelect.setTimeOut(30);
// while(1) {
//		mySelect.check();
//		while( cptr = mySelect.nextReadPtr) ) {
//			...
//		}
// }
///////////////////////////////////////////////////////////////////////////
// Another usage would be to handle UpSockets (or classes containing UpSockets)
//
// UpSelect<UpSocket*> sockSelect;
// UpSocket sock1,sock2,*sptr;
// ...
// sockSelect.set(sock1.fd(), &sock1, true, true, false );
// sockSelect.set(sock2.fd(), &sock2, true, true, false );
// sockeSelect.setTimeoutIndefinite();
// while(1) {
//		sockSelect.check();
//		while( sptr = mySelect.nextReadPtr() ) {
//			...
//		}
//		while( sptr = mySelect.nextWritePtr() ) {
//			...
//		}
// }

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

#ifndef UPSELECT_H
#define UPSELECT_H

// unix includes
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
// c++ includes
#include <map>
#include <cstring>
// UpTools Includes
#include <UpTools/UpTimeVal.h>

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// template UpSelect<ChannelPointerClass>

// Tptr should be a pointer or a smart pointer to a channel class
// it defaults to void* if not specified
template<typename Tptr=void*>
class UpSelect {
private:
		  
	// Tptr container indexed by file descr.
	typedef std::map<int,Tptr> PtrMap; 
	typedef typename PtrMap::iterator   PtrMapIterator;
	typedef typename PtrMap::value_type PtrMapValueType;
	
	UpTimeVal	timeout;
	bool			indefinite;
	timeval		timer,*timerPtr;
	
	fd_set		readFdset, writeFdset, exceptFdset;
	fd_set		rFdset,    wFdset,     eFdset;
	int			maxFd;
	
	PtrMap			channelMap, readMap, writeMap, exceptMap;
	PtrMapIterator	channelIt,  readIt,  writeIt,  exceptIt,  auxIt;

	std::string	lastError;

public:
	UpSelect();
	~UpSelect();

	// return number of configured channels
	unsigned int count();				// total number of configured channels
	unsigned int readCount();			// total number of configured read channels
	unsigned int writeCount();			// total number of configured write channels
	unsigned int exceptionCount();	// total number of conf exception channels

	// add channels for event checking (read and/or write and/or exception)
	// returns true if channel is succesfully inserted or modified
	// channels ar indexed by file descriptor fd
	bool set(int fd, Tptr s,bool _read, bool _write, bool _exception );
	
	// remove channel from event checking
	bool remove(int fd);
	// remove all channels
	void removeAll();

	// makes check() to wait a time t for an event to arrive
	// (if no timeout is set, the default timer is Indefinite,
	// that is, wait for an event for ever)
	void setTimeOut(time_t t);
	void setTimeOut(const timeval& t);
	void setTimeOut(const UpTimeVal& t);
	void setTimeOut(double t);
	// makes check() to wait forever until an event happens
	void setTimeoutIndefinite();
	

	// the following method checks if any of the added channels is ready
	// returns the number of events registered or:
	// 	0 if a timeout ocurred
	// 	-1 if an error ocurred 
	// the timer is reset every time this function is called
	int check();

	// report last error
	const std::string& getLastError();

	// the following return a pointer to a stored channel
	// they return a NULL if there are no more channels available
	// they are used to see what channels are stored
	Tptr firstPtr();
	Tptr nextPtr();
	// find a channel
	Tptr findPtr(int _fd);

	// the following return a pointer to the next channel available after a check
	// they return a NULL if there are no more channels available
	Tptr nextReadPtr();
	Tptr nextWritePtr();
	Tptr nextExceptionPtr();

	// the following return a file descriptor to the next channel available
	// after a check
	// they return -1 if there are no more channels available
	int nextRead();
	int nextWrite();
	int nextException();

};


////////////////////////////////////////////////////////////////////////
/* from man select_tut
       readfds
              This set is watched to see if data is available for reading from
              any of its file descriptors. After select has returned,  readfds
              will  be  cleared  of all file descriptors except for those file
              descriptors that are immediately available for  reading  with  a
              recv()  (for  sockets) or read() (for pipes, files, and sockets)
              call.

       writefds
              This set is watched to see if there is space to  write  data  to
              any  of its file descriptor. After select has returned, writefds
              will be cleared of all file descriptors except  for  those  file
              descriptors  that  are  immediately available for writing with a
              send() (for sockets) or write() (for pipes, files, and  sockets)
              call.

       exceptfds
              This  set is watched for exceptions or errors on any of the file
              descriptors. However, that is actually just a rumor. How you use
              exceptfds  is  to  watch for out-of-band (OOB) data. OOB data is
              data sent  on  a  socket  using  the  MSG_OOB  flag,  and  hence
              exceptfds  only  really  applies  to  sockets.  See  recv(2) and
              send(2) about this. After select has returned, exceptfds will be
              cleared  of  all  file  descriptors except for those descriptors
              that are available for reading OOB data. You can only ever  read
              one  byte  of  OOB  data though (which is done with recv()), and
              writing OOB data (done with send) can be done at  any  time  and
              will not block. Hence there is no need for a fourth set to check
              if a socket is available for writing OOB data.

*/

	
////////////////////////////////////////////////////////////////////////	

/* from man fcntl:

       F_SETFL
              Set the file status flags part of the descriptor's flags to  the
              value  specified by arg.  Remaining bits (access mode, file cre-
              ation flags) in arg are ignored.  On Linux this command can only
              change the O_APPEND, O_NONBLOCK, O_ASYNC, and O_DIRECT flags.

	O_APPEND is not applicable for sockets.

	from man 2 open:
       O_APPEND
              The  file  is opened in append mode. Before each write, the file
              pointer is positioned at the end of the file, as if with  lseek.
              O_APPEND may lead to corrupted files on NFS file systems if more
              than one process appends data  to  a  file  at  once.   This  is
              because  NFS does not support appending to a file, so the client
              kernel has to simulate it, which can't be done  without  a  race
              condition.
       O_NONBLOCK or O_NDELAY
              When  possible, the file is opened in non-blocking mode. Neither
              the open nor any subsequent operations on  the  file  descriptor
              which  is  returned will cause the calling process to wait.  For
              the handling of FIFOs (named pipes),  see  also  fifo(4).   This
              mode need not have any effect on files other than FIFOs.
        O_DIRECT
              Try  to minimize cache effects of the I/O to and from this file.
              In general this will degrade performance, but it  is  useful  in
              special  situations,  such  as  when  applications  do their own
              caching.  File I/O is done directly to/from user space  buffers.
              The  I/O  is synchronous, i.e., at the completion of the read(2)
              or write(2) system call, data is guaranteed to have been  trans-
              ferred.   Transfer  sizes,  and the alignment of user buffer and
              file offset must all be multiples of the logical block  size  of
              the file system.
              This flag is supported on a number of Unix-like systems; support
              was added under Linux in kernel version 2.4.10.
              A semantically similar interface for block devices is  described
              in raw(8).
       O_ASYNC
              Generate a signal (SIGIO by default, but this can be changed via
              fcntl(2)) when input or output becomes  possible  on  this  file
              descriptor.   This  feature  is  only  available  for terminals,
              pseudo-terminals, and sockets. See fcntl(2) for further details.
*/

///////////////////////////////////////////////////////////////////////////
// inline code

#include<cerrno>

#define MAX(a,b) ((a<b)?(b):(a))

template<typename Tptr>
inline UpSelect<Tptr>::UpSelect()
: timeout(0,0), indefinite(true), maxFd(0)
{
	channelIt=channelMap.begin();
	FD_ZERO(&readFdset);		FD_ZERO(&rFdset); readIt   = readMap.begin();
	FD_ZERO(&writeFdset);	FD_ZERO(&wFdset); writeIt  = writeMap.begin();
	FD_ZERO(&exceptFdset);	FD_ZERO(&eFdset); exceptIt = exceptMap.begin();
}

template<typename Tptr>
inline UpSelect<Tptr>::~UpSelect()
{ }

template<typename Tptr>
inline void UpSelect<Tptr>::setTimeOut(time_t t)
{ timeout = t; indefinite=false; }

template<typename Tptr>
inline void UpSelect<Tptr>::setTimeOut(const timeval& t)
{ timeout = t; indefinite=false; }

template<typename Tptr>
inline void UpSelect<Tptr>::setTimeOut(const UpTimeVal& t)
{ timeout = t; indefinite=false; }

template<typename Tptr>
inline void UpSelect<Tptr>::setTimeOut(double t)
{ timeout = t; indefinite=false; }

template<typename Tptr>
inline void UpSelect<Tptr>::setTimeoutIndefinite()
{ indefinite =true; }// wait forever

template<typename Tptr>
inline const std::string& UpSelect<Tptr>::getLastError()
{ return lastError; }

template<typename Tptr>
inline unsigned int UpSelect<Tptr>::count()
{ return channelMap.size(); }

template<typename Tptr>
inline unsigned int UpSelect<Tptr>::readCount()
{ return readMap.size(); }

template<typename Tptr>
inline unsigned int UpSelect<Tptr>::writeCount()
{ return writeMap.size(); }

template<typename Tptr>
inline unsigned int UpSelect<Tptr>::exceptionCount()
{ return exceptMap.size(); }


template<typename Tptr>
inline Tptr UpSelect<Tptr>::firstPtr()
{ return channelMap.size() ? (channelIt=channelMap.begin())->second : 0 ; }

template<typename Tptr>
inline Tptr UpSelect<Tptr>::nextPtr()
{ channelIt++; return (channelIt!=channelMap.end()) ? channelIt->second : 0 ; }

template<typename Tptr>
inline Tptr UpSelect<Tptr>::findPtr(int _fd)
{
	PtrMapIterator it=channelMap.find(_fd);
	return (it!=channelMap.end()) ? it->second : 0;
}


template<typename Tptr>
inline bool UpSelect<Tptr>::set(int fd, Tptr s, bool _read, bool _write, bool _except )
{
  std::pair<PtrMapIterator,bool>
			ir( channelMap.insert( PtrMapValueType(fd,s) ) );

  if( ! ir.second ) { // file descriptor was already present
		if( ir.first->second != s ) { // same fd but different pointer
			return false;
		}
	}

	register	bool priorRule = FD_ISSET(fd,&readFdset);
	if( _read != priorRule ) {
		if( _read ) { // must check for read-ok events in fd
				FD_SET(fd,&readFdset); readMap[fd]=s;
		} else { // must not check for read-ok events in fd
				if( readIt!=readMap.end() && readIt->first==fd ) readIt++;
				FD_CLR(fd,&readFdset); readMap.erase(fd);
		}
	}

	priorRule = FD_ISSET(fd,&writeFdset);
	if( _write != priorRule ) {
		if( _write ) { // must check for write-ok events in fd
			FD_SET(fd,&writeFdset); writeMap[fd]=s;
		} else { // must not check for write-ok events in fd
			if( writeIt!=writeMap.end() && writeIt->first==fd ) writeIt++;
			FD_CLR(fd,&writeFdset); writeMap.erase(fd);
		}
	}

	priorRule = FD_ISSET(fd,&exceptFdset);
	if( _except != priorRule ) {
		if( _except ) { // must check for out-of-band or urgent data in fd
			FD_SET(fd,&exceptFdset); exceptMap[fd]=s;
		} else { // must not check for out-of-band or urgent data in fd
			if( exceptIt!=exceptMap.end() && exceptIt->first==fd ) exceptIt++;
			FD_CLR(fd,&exceptFdset); exceptMap.erase(fd);
		}
	}

	maxFd = MAX(maxFd-1,fd)+1;
	return true;
}


template<typename Tptr>
inline void UpSelect<Tptr>::removeAll()
{
	maxFd=0;
	channelMap.clear();		channelIt=channelMap.begin();
	FD_ZERO(&readFdset);		FD_ZERO(&rFdset);
	readMap.clear();			readIt = readMap.begin();
	FD_ZERO(&writeFdset);	FD_ZERO(&wFdset);
	writeMap.clear();			writeIt  = writeMap.begin();
	FD_ZERO(&exceptFdset);	FD_ZERO(&eFdset);
	exceptMap.clear();		exceptIt = exceptMap.begin();
}

template<typename Tptr>
inline bool UpSelect<Tptr>::remove(int fd)
{
	bool ret;
	if( channelIt!=channelMap.end() && channelIt->first==fd )
		{ auxIt = channelIt; channelIt++; channelMap.erase(auxIt); ret = true; }
	else { ret = ( channelMap.erase(fd) != 0 ); }

	if( FD_ISSET(fd,&readFdset) ) {
		FD_CLR(fd,&readFdset);
		if( readIt!=readMap.end() && readIt->first==fd )
			{ auxIt=readIt; readIt++; readMap.erase(auxIt); }
		else { readMap.erase(fd); }
	}

	if( FD_ISSET(fd,&writeFdset) ) {
		FD_CLR(fd,&writeFdset);
		if( writeIt!=writeMap.end() && writeIt->first==fd )
			{ auxIt=writeIt; writeIt++; writeMap.erase(auxIt); }
		else { writeMap.erase(fd); }
	}

	if( FD_ISSET(fd,&exceptFdset) ) {
		FD_CLR(fd,&exceptFdset);
		if( exceptIt!=exceptMap.end() && exceptIt->first==fd )
			{ auxIt=exceptIt; exceptIt++; exceptMap.erase(auxIt); }
		else { exceptMap.erase(fd); }
	}

	maxFd=0;
	if(channelMap.size()) {
		auxIt = channelMap.end(); auxIt--;
		maxFd = auxIt->first + 1;
	}

	return ret; // returns removed channel ptr
}


template<typename Tptr>
inline int UpSelect<Tptr>::check()
{
	if(indefinite) {
		timerPtr = 0;
	} else {
		timer.tv_sec  = timeout.tv_sec;
		timer.tv_usec = timeout.tv_usec;
		timerPtr = &timer;
	}

	std::memcpy( &rFdset, &readFdset,   sizeof(fd_set) );
	std::memcpy( &wFdset, &writeFdset,  sizeof(fd_set) );
	std::memcpy( &eFdset, &exceptFdset, sizeof(fd_set) );

	readIt   = readMap.begin();
	writeIt  = writeMap.begin();
	exceptIt = exceptMap.begin();

	int n = select(maxFd,&rFdset,&wFdset,&eFdset,timerPtr);
	
	if(n<0) lastError.assign( std::strerror(errno) );
	else    lastError.clear();

	return n;
}



template<typename Tptr>
inline Tptr UpSelect<Tptr>::nextReadPtr()
{
	while( readIt != readMap.end() ) {
		auxIt = readIt;
		readIt++;
		if( FD_ISSET( auxIt->first , &rFdset ) ) return auxIt->second;
	}
	return 0;
}

template<typename Tptr>
inline int UpSelect<Tptr>::nextRead()
{
	while( readIt != readMap.end() ) {
		auxIt = readIt;
		readIt++;
		if( FD_ISSET( auxIt->first , &rFdset ) ) return auxIt->first;
	}
	return -1;
}

template<typename Tptr>
inline Tptr UpSelect<Tptr>::nextWritePtr()
{
	while( writeIt != writeMap.end() ) {
		auxIt = writeIt;
		writeIt++;
		if( FD_ISSET( auxIt->first , &wFdset ) ) return auxIt->second;
	}
	return 0;
}

template<typename Tptr>
inline int UpSelect<Tptr>::nextWrite()
{
	while( writeIt != writeMap.end() ) {
		auxIt = writeIt;
		writeIt++;
		if( FD_ISSET( auxIt->first , &wFdset ) ) return auxIt->first;
	}
	return -1;
}

template<typename Tptr>
inline Tptr UpSelect<Tptr>::nextExceptionPtr()
{
	while( exceptIt != exceptMap.end() ) {
		auxIt = exceptIt;
		exceptIt++;
		if( FD_ISSET( auxIt->first , &eFdset ) ) return auxIt->second;
	}
	return 0;
}

template<typename Tptr>
inline int UpSelect<Tptr>::nextException()
{
	while( exceptIt != exceptMap.end() ) {
		auxIt = exceptIt;
		exceptIt++;
		if( FD_ISSET( auxIt->first , &eFdset ) ) return auxIt->first;
	}
	return -1;
}



#endif
