/* UpTools v8.6
 *
 * Copyright (c) 2005-2011 Fundacion Universidad de Palermo (Argentina).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment: 'This product includes software developed by the
 *    "Universidad de Palermo, Argentina" (http://www.palermo.edu/).'
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef UPTHEADSPECIFIC_INL
#define UPTHEADSPECIFIC_INL

#include <cassert>
/////////////////////////////////////////////////////////////////////////////
// UpThreadSpecificPtr methods //////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

template<typename T>
inline UpThreadSpecificPtr<T>::UpThreadSpecificPtr()
{ assert( pthread_key_create(&k,0)==0 ); }

template<typename T>
inline UpThreadSpecificPtr<T>::UpThreadSpecificPtr(const T* ptr)
{ assert( pthread_key_create(&k,0)==0 ); set(ptr); }

template<typename T>
inline UpThreadSpecificPtr<T>::UpThreadSpecificPtr(const UpThreadSpecificPtr<T>& sptr)
{ assert( pthread_key_create(&k,0)==0 ); set(sptr.get()); } 

template<typename T>
inline UpThreadSpecificPtr<T>::~UpThreadSpecificPtr()
{ pthread_key_delete(k); }

template<typename T>
inline T* UpThreadSpecificPtr<T>::get() const
{ return (T*)pthread_getspecific(k); }

template<typename T>
inline void UpThreadSpecificPtr<T>::set(const T* ptr)
{ pthread_setspecific(k,ptr); }

// operadores de asignacion
template<typename T>
inline UpThreadSpecificPtr<T>& UpThreadSpecificPtr<T>::operator=(const T* ptr)
{ set(ptr); return *this; }

template<typename T>
inline UpThreadSpecificPtr<T>& UpThreadSpecificPtr<T>::operator=(const UpThreadSpecificPtr<T>& sptr)
{ set(sptr.get()); return *this; }

// cast to void* operator
template<typename T>
inline UpThreadSpecificPtr<T>::operator void*()
{ return get(); } 

// cast to void* operator
template<typename T>
inline UpThreadSpecificPtr<T>::operator T*()
{ return get(); } 

// operadores utiles
template<typename T>
inline T& UpThreadSpecificPtr<T>::operator*()
{ return *get(); }

template<typename T>
inline const T& UpThreadSpecificPtr<T>::operator*() const
{ return *get(); }

template<typename T>
inline T* UpThreadSpecificPtr<T>::operator->()
{ return get(); }

template<typename T>
inline const T* UpThreadSpecificPtr<T>::operator->() const
{ return get(); }

template<typename T>
inline bool	UpThreadSpecificPtr<T>::operator!() const
{ return get() ? false : true; }

// operadores de comparacion
template<typename T>
inline bool	UpThreadSpecificPtr<T>::operator==(const T* ptr) const
{ return get()==(T*)ptr; }  

template<typename T>
inline bool	UpThreadSpecificPtr<T>::operator!=(const T* ptr) const
{ return get()!=(T*)ptr; }

template<typename T>
inline bool	UpThreadSpecificPtr<T>::operator==(const UpThreadSpecificPtr<T>& sptr) const
{ return get()==(T*)(sptr.get()); }

template<typename T>
inline bool	UpThreadSpecificPtr<T>::operator!=(const UpThreadSpecificPtr<T>& sptr) const
{ return get()!=(T*)(sptr.get()); }

template<typename T>
inline bool UpThreadSpecificPtr<T>::operator<(const UpThreadSpecificPtr<T>& sptr) const
{ return get()<(T*)(sptr.get()); }

/////////////////////////////////////////////////////////////////////////////
// UpThreadSpecificSmartPtr methods /////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

template<typename T>
inline UpThreadSpecificSmartPtr<T>::UpThreadSpecificSmartPtr()
{ assert( pthread_key_create(&k,0)==0 ); }

template<typename T>
inline UpThreadSpecificSmartPtr<T>::UpThreadSpecificSmartPtr(const T* ptr) {
  assert( pthread_key_create(&k,0)==0 );
  hold(ptr); if(ptr) set(ptr); }

template<typename T>
inline UpThreadSpecificSmartPtr<T>::UpThreadSpecificSmartPtr(const UpThreadSpecificSmartPtr<T>& sptr) {
  assert( pthread_key_create(&k,0)==0 ); T* ptr(sptr.get());
  hold(ptr); if(ptr) set(ptr);
} 

template<typename T>
inline UpThreadSpecificSmartPtr<T>::~UpThreadSpecificSmartPtr()
{ T* ptr(get()); pthread_key_delete(k); release(ptr); }

template<typename T>
inline T* UpThreadSpecificSmartPtr<T>::get() const
{ return (T*)pthread_getspecific(k); }

template<typename T>
inline void UpThreadSpecificSmartPtr<T>::set(const T* ptr)
{ pthread_setspecific(k,ptr); }

#include <UpTools/UpSmartLinkPtr.h> // included here for the macro definitions, will be removed later

template<typename T>
inline void UpThreadSpecificSmartPtr<T>::hold(const T* sp) {
   if(sp) {
      sp->link();
      MONITOR_LINK_EVENT
   }
}

template<typename T>
inline void UpThreadSpecificSmartPtr<T>::release(const T* sp) {
   if(sp) {
      #ifdef UPLINKABLE_LOG_LINKS
      int count=sp->unlink();
      MONITOR_UNLINK_EVENT
      if( count==0 ) delete sp;
      #else
      if( sp->unlinkAndReturnPriorValue()==1 )
         delete sp;
      #endif
   }
}

// operadores de asignacion
template<typename T>
inline UpThreadSpecificSmartPtr<T>& UpThreadSpecificSmartPtr<T>::operator=(const T* newptr)
{ T *oldptr(get());
  if( oldptr==newptr ) return *this;
  hold(newptr); release(oldptr); set(newptr);
  return this;
}

template<typename T>
inline UpThreadSpecificSmartPtr<T>& UpThreadSpecificSmartPtr<T>::operator=(const UpThreadSpecificSmartPtr<T>& sptr)
{ T *oldptr(get()), *newptr(sptr.get());
  if( oldptr==newptr ) return *this;
  hold(newptr); release(oldptr); set(newptr);
  return this;
}

// cast to void* operator
template<typename T>
inline UpThreadSpecificSmartPtr<T>::operator void*()
{ return get(); } 

// cast to T* operator
template<typename T>
inline UpThreadSpecificSmartPtr<T>::operator T*()
{ return get(); } 

// operadores utiles
template<typename T>
inline T& UpThreadSpecificSmartPtr<T>::operator*()
{ return *get(); }

template<typename T>
inline const T& UpThreadSpecificSmartPtr<T>::operator*() const
{ return *get(); }

template<typename T>
inline T* UpThreadSpecificSmartPtr<T>::operator->()
{ return get(); }

template<typename T>
inline const T* UpThreadSpecificSmartPtr<T>::operator->() const
{ return get(); }

template<typename T>
inline bool	UpThreadSpecificSmartPtr<T>::operator!() const
{ return get() ? false : true; }

// operadores de comparacion
template<typename T>
inline bool	UpThreadSpecificSmartPtr<T>::operator==(const T* ptr) const
{ return get()==(T*)ptr; }  

template<typename T>
inline bool	UpThreadSpecificSmartPtr<T>::operator!=(const T* ptr) const
{ return get()!=(T*)ptr; }

template<typename T>
inline bool	UpThreadSpecificSmartPtr<T>::operator==(const UpThreadSpecificSmartPtr<T>& sptr) const
{ return get()==(T*)(sptr.get()); }

template<typename T>
inline bool	UpThreadSpecificSmartPtr<T>::operator!=(const UpThreadSpecificSmartPtr<T>& sptr) const
{ return get()!=(T*)(sptr.get()); }

template<typename T>
inline bool UpThreadSpecificSmartPtr<T>::operator<(const UpThreadSpecificSmartPtr<T>& sptr) const
{ return get()<(T*)(sptr.get()); }

#endif
